@@ -125,11 +125,20 @@ int inet_ptom (int af, const char *src, unsigned int *dst)
  */
 const char *inet_mtop (int af, unsigned int src, char *dst, socklen_t len)
 {
-    struct in_addr addr;
-    
-    inet_mton(AF_INET, src, &addr);
-    
-    return inet_ntop (af, &addr, dst, len);
+    if (af == AF_INET) {
+        struct in_addr addr;
+
+        inet_mton(af, src, &addr);
+        return inet_ntop(af, &addr, dst, len);
+    } else if (af == AF_INET6) {
+        struct in6_addr *addr6;
+
+        inet_mton(af, src, &addr6);
+        return inet_ntop(af, &addr6, dst, len);
+    } else {
+        di_warning("Unknown address family passed to inet_mtop: %i", af);
+        return NULL;
+    }
 }
 
 /* convert a mask length (eg 24) in +src+ into the struct in_addr it corresponds
@@ -245,7 +254,8 @@ int is_raw_80211(const char *iface)
     struct ifreq ifr;
     struct sockaddr sa;
 
-    strncpy(ifr.ifr_name, iface, IFNAMSIZ);
+    strncpy(ifr.ifr_name, iface, IFNAMSIZ-1);
+    ifr.ifr_name[IFNAMSIZ-1]='\0';
 
     if (skfd && ioctl(skfd, SIOCGIFHWADDR, &ifr) < 0) {
         di_warning("Unable to retrieve interface type.");
@@ -293,11 +303,11 @@ int is_layer3_qeth(const char *iface)
         goto out;
     }
 
-    buf[slen + 1] = '\0';
+    buf[slen] = '\0';
 
     driver = strrchr(buf, '/') + 1;
     if (strcmp(driver, "qeth") != 0) {
-        di_error("no qeth found: %s", driver);
+        di_info("no qeth found: %s", driver);
         goto out;
     }
 
@@ -408,12 +418,13 @@ int get_all_ifs (int all, char*** ptr)
     char ibuf[512];
     char** list = NULL;
     size_t len = 0;
+    ibuf[sizeof(ibuf)-1]='\0';
 
     if (getifaddrs(&ifap) == -1)
         return 0;
 
     for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
-        strncpy(ibuf, ifa->ifa_name, sizeof(ibuf));
+        strncpy(ibuf, ifa->ifa_name, sizeof(ibuf)-1);
         if (ifa->ifa_flags & IFF_LOOPBACK)   /* ignore loopback devices */
             continue;
 #if defined(__linux__)
@@ -1310,11 +1321,12 @@ void interface_up (const char *if_name)
 {
     struct ifreq ifr;
 
-    strncpy(ifr.ifr_name, if_name, IFNAMSIZ);
+    strncpy(ifr.ifr_name, if_name, IFNAMSIZ-1);
+    ifr.ifr_name[IFNAMSIZ-1]='\0';
 
     if (skfd && ioctl(skfd, SIOCGIFFLAGS, &ifr) >= 0) {
         di_info("Activating interface %s", if_name);
-        strncpy(ifr.ifr_name, if_name, IFNAMSIZ);
+        strncpy(ifr.ifr_name, if_name, IFNAMSIZ-1);
         ifr.ifr_flags |= (IFF_UP | IFF_RUNNING);
         ioctl(skfd, SIOCSIFFLAGS, &ifr);
     } else {
@@ -1326,11 +1338,12 @@ void interface_down (const char *if_name)
 {
     struct ifreq ifr;
 
-    strncpy(ifr.ifr_name, if_name, IFNAMSIZ);
+    strncpy(ifr.ifr_name, if_name, IFNAMSIZ-1);
+    ifr.ifr_name[IFNAMSIZ-1]='\0';
 
     if (skfd && ioctl(skfd, SIOCGIFFLAGS, &ifr) >= 0) {
         di_info("Taking down interface %s", if_name);
-        strncpy(ifr.ifr_name, if_name, IFNAMSIZ);
+        strncpy(ifr.ifr_name, if_name, IFNAMSIZ-1);
         ifr.ifr_flags &= ~IFF_UP;
         ioctl(skfd, SIOCSIFFLAGS, &ifr);
     } else {
@@ -1371,11 +1384,12 @@ void reap_old_files (void)
 /* Convert a space-separated list of nameservers in a single string (as might
  * be entered into, say, debconf), and store them in the interface.
  */
-void netcfg_nameservers_to_array(const char *nameservers, struct netcfg_interface *interface)
+int netcfg_nameservers_to_array(const char *nameservers, struct netcfg_interface *interface)
 {
     char *save, *ptr, *ns;
     unsigned int i;
     union inX_addr addr;
+    int ok = 1;
 
     if (nameservers) {
         save = ptr = strdup(nameservers);
@@ -1398,7 +1412,8 @@ void netcfg_nameservers_to_array(const char *nameservers, struct netcfg_interfac
                         af = AF_INET6;
                     } else {
                         af = -1;
-                        fprintf(stderr, "Failed to parse %s as an IP address", ns);
+                        fprintf(stderr, "Failed to parse %s as an IP address\n", ns);
+                        ok = 0;
                     }
                 }
 
@@ -1417,12 +1432,14 @@ void netcfg_nameservers_to_array(const char *nameservers, struct netcfg_interfac
         /* Empty out all the nameserver strings */
         for (i = 0; i < NETCFG_NAMESERVERS_MAX; i++) *(interface->nameservers[i]) = '\0';
     }
+    return ok;
 }
 
-int netcfg_get_nameservers (struct debconfclient *client, char **nameservers, char *default_nameservers)
+int netcfg_get_nameservers (struct debconfclient *client, char **nameservers, char *default_nameservers,
+                            struct netcfg_interface *interface)
 {
     char *ptr;
-    int ret;
+    int ret, ok = 0;
 
     debconf_get(client,"netcfg/get_nameservers");
     if (*nameservers)
@@ -1435,14 +1452,24 @@ int netcfg_get_nameservers (struct debconfclient *client, char **nameservers, ch
         ptr = "";
     debconf_set(client, "netcfg/get_nameservers", ptr);
 
-    debconf_input(client, "critical", "netcfg/get_nameservers");
-    ret = debconf_go(client);
+    while (!ok) {
+        debconf_input(client, "critical", "netcfg/get_nameservers");
+        ret = debconf_go(client);
 
-    if (ret)
-        return ret;
+        if (ret)
+            return ret;
 
-    debconf_get(client, "netcfg/get_nameservers");
-    ptr = client->value;
+        debconf_get(client, "netcfg/get_nameservers");
+        ptr = client->value;
+        ok = netcfg_nameservers_to_array(ptr, interface);
+
+        if (!ok) {
+            debconf_capb(client);
+            debconf_input (client, "critical", "netcfg/bad_ipaddress");
+            debconf_go (client);
+            debconf_capb(client, "backup");
+        }
+    }
 
     if (*nameservers)
         free(*nameservers);
@@ -1469,7 +1496,7 @@ int netcfg_detect_link(struct debconfclient *client, const struct netcfg_interfa
 {
     char arping[256];
     int count, rv = 0;
-    int link_waits;
+    int link_waits = 12;  /* default for netcfg/link_wait_timeout times 4 */
     int gw_tries = NETCFG_GATEWAY_REACHABILITY_TRIES;
     const char *if_name = interface->name;
     const char *gateway = interface->gateway;
@@ -1577,10 +1604,12 @@ int netcfg_parse_cidr_address(const char *address, struct netcfg_interface *inte
     struct in_addr addr;
     struct in6_addr addr6;
     int ok;
-    char *maskptr, addrstr[NETCFG_ADDRSTRLEN];
+    char *maskptr, *addrstr, addrbuf[NETCFG_ADDRSTRLEN];
     int i;
     
-    strncpy(addrstr, address, NETCFG_ADDRSTRLEN);
+    strncpy(addrbuf, address, NETCFG_ADDRSTRLEN-1);
+    addrbuf[NETCFG_ADDRSTRLEN-1]='\0';
+    addrstr = strtrim(addrbuf);
     
     if ((maskptr = strchr(addrstr, '/'))) {
         /* Houston, we have a netmask; split it into bits */
@@ -1674,10 +1703,27 @@ int netcfg_gateway_reachable(const struct netcfg_interface *interface)
     inet_pton(interface->address_family, interface->gateway, &gw_addr);
 
     if (interface->address_family == AF_INET) {
+        if (interface->masklen == 32) {
+            /* Special case for single-address networks.  We'll handle
+             * routing manually in this case.
+             */
+            return 1;
+        }
+
         return (gw_addr.in4.s_addr && ((gw_addr.in4.s_addr & mask.in4.s_addr) == net.in4.s_addr));
     } else if (interface->address_family == AF_INET6) {
+        if (interface->masklen == 128) {
+            /* Special case for single-address networks.  We'll handle
+             * routing manually in this case.
+             */
+            return 1;
+        }
+
+        if ((ntohs(gw_addr.in6.s6_addr16[0]) & 0xffc0) == (0xfe80 & 0xffc0)) {
+            return 1;
+        }
+
         int i;
-        
         for (i = 0; i < 4; i++) {
             if ((gw_addr.in6.s6_addr32[i] & mask.in6.s6_addr32[i]) != net.in6.s6_addr32[i]) {
                 return 0;
@@ -1730,7 +1776,24 @@ void rtrim(char *s)
 	
 	n = strlen(s) - 1;
 	
-	while (isspace(s[n])) {
+	while (n >= 0 && isspace(s[n])) {
 		s[n] = '\0';
+		n--;
 	}
 }
+
+char *strtrim(char *s)
+{
+	size_t len;
+
+	len = strlen(s);
+	if (!len)
+		return s;
+
+	rtrim(s);
+
+	while (*s && isspace(*s))
+		s++;
+
+	return s;
+}
